# ------------------------------------------------------------------------
# Deformable DETR
# Copyright (c) 2020 SenseTime. All Rights Reserved.
# Licensed under the Apache License, Version 2.0 [see LICENSE for details]
# ------------------------------------------------------------------------
# Modified from DETR (https://github.com/facebookresearch/detr)
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
# ------------------------------------------------------------------------
from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
import copy
from typing import Optional, List
import math
import torch
import torch.nn.functional as F
from torch import nn, Tensor
from torch.nn.init import xavier_uniform_, constant_, uniform_, normal_
from torch.autograd import Function
from torch.autograd.function import once_differentiable
import MultiScaleDeformableAttention as MSDA
import warnings
import math
from torch import nn


def inverse_sigmoid(x, eps=1e-5):
    x = x.clamp(min=0, max=1)
    x1 = x.clamp(min=eps)
    x2 = (1 - x).clamp(min=eps)
    return torch.log(x1 / x2)


class MSDeformAttnFunction(Function):
    @staticmethod
    def forward(
        ctx,
        value,
        value_spatial_shapes,
        value_level_start_index,
        sampling_locations,
        attention_weights,
        im2col_step,
    ):
        ctx.im2col_step = im2col_step
        output = MSDA.ms_deform_attn_forward(
            value,
            value_spatial_shapes,
            value_level_start_index,
            sampling_locations,
            attention_weights,
            ctx.im2col_step,
        )
        ctx.save_for_backward(
            value,
            value_spatial_shapes,
            value_level_start_index,
            sampling_locations,
            attention_weights,
        )
        return output

    @staticmethod
    @once_differentiable
    def backward(ctx, grad_output):
        (
            value,
            value_spatial_shapes,
            value_level_start_index,
            sampling_locations,
            attention_weights,
        ) = ctx.saved_tensors
        grad_value, grad_sampling_loc, grad_attn_weight = MSDA.ms_deform_attn_backward(
            value,
            value_spatial_shapes,
            value_level_start_index,
            sampling_locations,
            attention_weights,
            grad_output,
            ctx.im2col_step,
        )

        return grad_value, None, None, grad_sampling_loc, grad_attn_weight, None


def ms_deform_attn_core_pytorch(
    value, value_spatial_shapes, sampling_locations, attention_weights
):
    # for debug and test only,
    # need to use cuda version instead
    N_, S_, M_, D_ = value.shape
    _, Lq_, M_, L_, P_, _ = sampling_locations.shape
    value_list = value.split([H_ * W_ for H_, W_ in value_spatial_shapes], dim=1)
    sampling_grids = 2 * sampling_locations - 1
    sampling_value_list = []
    for lid_, (H_, W_) in enumerate(value_spatial_shapes):
        # N_, H_*W_, M_, D_ -> N_, H_*W_, M_*D_ -> N_, M_*D_, H_*W_ -> N_*M_, D_, H_, W_
        value_l_ = (
            value_list[lid_].flatten(2).transpose(1, 2).reshape(N_ * M_, D_, H_, W_)
        )
        # N_, Lq_, M_, P_, 2 -> N_, M_, Lq_, P_, 2 -> N_*M_, Lq_, P_, 2
        sampling_grid_l_ = sampling_grids[:, :, :, lid_].transpose(1, 2).flatten(0, 1)
        # N_*M_, D_, Lq_, P_
        sampling_value_l_ = F.grid_sample(
            value_l_,
            sampling_grid_l_,
            mode="bilinear",
            padding_mode="zeros",
            align_corners=False,
        )
        sampling_value_list.append(sampling_value_l_)
    # (N_, Lq_, M_, L_, P_) -> (N_, M_, Lq_, L_, P_) -> (N_, M_, 1, Lq_, L_*P_)
    attention_weights = attention_weights.transpose(1, 2).reshape(
        N_ * M_, 1, Lq_, L_ * P_
    )
    output = (
        (torch.stack(sampling_value_list, dim=-2).flatten(-2) * attention_weights)
        .sum(-1)
        .view(N_, M_ * D_, Lq_)
    )
    return output.transpose(1, 2).contiguous()


def _is_power_of_2(n):
    if (not isinstance(n, int)) or (n < 0):
        raise ValueError(
            "invalid input for _is_power_of_2: {} (type: {})".format(n, type(n))
        )
    return (n & (n - 1) == 0) and n != 0


class MSDeformAttn(nn.Module):
    def __init__(self, d_model=256, n_levels=4, n_heads=8, n_points=4):
        """
        Multi-Scale Deformable Attention Module
        :param d_model      hidden dimension
        :param n_levels     number of feature levels
        :param n_heads      number of attention heads
        :param n_points     number of sampling points per attention head per feature level
        """
        super().__init__()
        if d_model % n_heads != 0:
            raise ValueError(
                "d_model must be divisible by n_heads, but got {} and {}".format(
                    d_model, n_heads
                )
            )
        _d_per_head = d_model // n_heads
        # you'd better set _d_per_head to a power of 2 which is more efficient in our CUDA implementation
        if not _is_power_of_2(_d_per_head):
            warnings.warn(
                "You'd better set d_model in MSDeformAttn to make the dimension of each attention head a power of 2 "
                "which is more efficient in our CUDA implementation."
            )

        self.im2col_step = 64

        self.d_model = d_model
        self.n_levels = n_levels
        self.n_heads = n_heads
        self.n_points = n_points

        self.sampling_offsets = nn.Linear(d_model, n_heads * n_levels * n_points * 2)
        self.attention_weights = nn.Linear(d_model, n_heads * n_levels * n_points)
        self.value_proj = nn.Linear(d_model, d_model)
        self.output_proj = nn.Linear(d_model, d_model)

        self._reset_parameters()

    def _reset_parameters(self):
        constant_(self.sampling_offsets.weight.data, 0.0)
        thetas = torch.arange(self.n_heads, dtype=torch.float32) * (
            2.0 * math.pi / self.n_heads
        )
        grid_init = torch.stack([thetas.cos(), thetas.sin()], -1)
        grid_init = (
            (grid_init / grid_init.abs().max(-1, keepdim=True)[0])
            .view(self.n_heads, 1, 1, 2)
            .repeat(1, self.n_levels, self.n_points, 1)
        )
        for i in range(self.n_points):
            grid_init[:, :, i, :] *= i + 1
        with torch.no_grad():
            self.sampling_offsets.bias = nn.Parameter(grid_init.view(-1))
        constant_(self.attention_weights.weight.data, 0.0)
        constant_(self.attention_weights.bias.data, 0.0)
        xavier_uniform_(self.value_proj.weight.data)
        constant_(self.value_proj.bias.data, 0.0)
        xavier_uniform_(self.output_proj.weight.data)
        constant_(self.output_proj.bias.data, 0.0)

    def forward(
        self,
        query,
        reference_points,
        input_flatten,
        input_spatial_shapes,
        input_level_start_index,
        input_padding_mask=None,
    ):
        """
        :param query                       (N, Length_{query}, C)
        :param reference_points            (N, Length_{query}, n_levels, 2), range in [0, 1], top-left (0,0), bottom-right (1, 1), including padding area
                                        or (N, Length_{query}, n_levels, 4), add additional (w, h) to form reference boxes
        :param input_flatten               (N, \sum_{l=0}^{L-1} H_l \cdot W_l, C)
        :param input_spatial_shapes        (n_levels, 2), [(H_0, W_0), (H_1, W_1), ..., (H_{L-1}, W_{L-1})]
        :param input_level_start_index     (n_levels, ), [0, H_0*W_0, H_0*W_0+H_1*W_1, H_0*W_0+H_1*W_1+H_2*W_2, ..., H_0*W_0+H_1*W_1+...+H_{L-1}*W_{L-1}]
        :param input_padding_mask          (N, \sum_{l=0}^{L-1} H_l \cdot W_l), True for padding elements, False for non-padding elements

        :return output                     (N, Length_{query}, C)
        """
        N, Len_q, _ = query.shape
        N, Len_in, _ = input_flatten.shape
        assert (input_spatial_shapes[:, 0] * input_spatial_shapes[:, 1]).sum() == Len_in

        value = self.value_proj(input_flatten)
        if input_padding_mask is not None:
            value = value.masked_fill(input_padding_mask[..., None], float(0))
        value = value.view(N, Len_in, self.n_heads, self.d_model // self.n_heads)
        sampling_offsets = self.sampling_offsets(query).view(
            N, Len_q, self.n_heads, self.n_levels, self.n_points, 2
        )
        attention_weights = self.attention_weights(query).view(
            N, Len_q, self.n_heads, self.n_levels * self.n_points
        )
        attention_weights = F.softmax(attention_weights, -1).view(
            N, Len_q, self.n_heads, self.n_levels, self.n_points
        )
        # N, Len_q, n_heads, n_levels, n_points, 2
        if reference_points.shape[-1] == 2:
            offset_normalizer = torch.stack(
                [input_spatial_shapes[..., 1], input_spatial_shapes[..., 0]], -1
            )
            sampling_locations = (
                reference_points[:, :, None, :, None, :]
                + sampling_offsets / offset_normalizer[None, None, None, :, None, :]
            )
        elif reference_points.shape[-1] == 4:
            sampling_locations = (
                reference_points[:, :, None, :, None, :2]
                + sampling_offsets
                / self.n_points
                * reference_points[:, :, None, :, None, 2:]
                * 0.5
            )
        else:
            raise ValueError(
                "Last dim of reference_points must be 2 or 4, but get {} instead.".format(
                    reference_points.shape[-1]
                )
            )
        output = MSDeformAttnFunction.apply(
            value,
            input_spatial_shapes,
            input_level_start_index,
            sampling_locations,
            attention_weights,
            self.im2col_step,
        )
        output = self.output_proj(output)
        return output


class DeformableTransformer(nn.Module):
    def __init__(
        self,
        d_model=256,
        nhead=8,
        num_encoder_layers=6,
        num_decoder_layers=6,
        dim_feedforward=1024,
        dropout=0.1,
        activation="relu",
        return_intermediate_dec=False,
        num_feature_levels=4,
        dec_n_points=4,
        enc_n_points=4,
        two_stage=False,
        two_stage_num_proposals=300,
    ):
        super().__init__()

        self.d_model = d_model
        self.nhead = nhead
        self.two_stage = two_stage
        self.two_stage_num_proposals = two_stage_num_proposals

        encoder_layer = DeformableTransformerEncoderLayer(
            d_model,
            dim_feedforward,
            dropout,
            activation,
            num_feature_levels,
            nhead,
            enc_n_points,
        )
        self.encoder = DeformableTransformerEncoder(encoder_layer, num_encoder_layers)

        decoder_layer = DeformableTransformerDecoderLayer(
            d_model,
            dim_feedforward,
            dropout,
            activation,
            num_feature_levels,
            nhead,
            dec_n_points,
        )
        self.decoder = DeformableTransformerDecoder(
            decoder_layer, num_decoder_layers, return_intermediate_dec
        )

        self.level_embed = nn.Parameter(torch.Tensor(num_feature_levels, d_model))

        if two_stage:
            self.enc_output = nn.Linear(d_model, d_model)
            self.enc_output_norm = nn.LayerNorm(d_model)
            self.pos_trans = nn.Linear(d_model * 2, d_model * 2)
            self.pos_trans_norm = nn.LayerNorm(d_model * 2)
        else:
            self.reference_points = nn.Linear(d_model, 2)

        self._reset_parameters()

    def _reset_parameters(self):
        for p in self.parameters():
            if p.dim() > 1:
                nn.init.xavier_uniform_(p)
        for m in self.modules():
            if isinstance(m, MSDeformAttn):
                m._reset_parameters()
        if not self.two_stage:
            xavier_uniform_(self.reference_points.weight.data, gain=1.0)
            constant_(self.reference_points.bias.data, 0.0)
        normal_(self.level_embed)

    def get_proposal_pos_embed(self, proposals):
        num_pos_feats = 128
        temperature = 10000
        scale = 2 * math.pi

        dim_t = torch.arange(
            num_pos_feats, dtype=torch.float32, device=proposals.device
        )
        dim_t = temperature ** (2 * (dim_t // 2) / num_pos_feats)
        # N, L, 4
        proposals = proposals.sigmoid() * scale
        # N, L, 4, 128
        pos = proposals[:, :, :, None] / dim_t
        # N, L, 4, 64, 2
        pos = torch.stack(
            (pos[:, :, :, 0::2].sin(), pos[:, :, :, 1::2].cos()), dim=4
        ).flatten(2)
        return pos

    def gen_encoder_output_proposals(self, memory, memory_padding_mask, spatial_shapes):
        N_, S_, C_ = memory.shape
        base_scale = 4.0
        proposals = []
        _cur = 0
        for lvl, (H_, W_) in enumerate(spatial_shapes):
            mask_flatten_ = memory_padding_mask[:, _cur : (_cur + H_ * W_)].view(
                N_, H_, W_, 1
            )
            valid_H = torch.sum(~mask_flatten_[:, :, 0, 0], 1)
            valid_W = torch.sum(~mask_flatten_[:, 0, :, 0], 1)

            grid_y, grid_x = torch.meshgrid(
                torch.linspace(
                    0, H_ - 1, H_, dtype=torch.float32, device=memory.device
                ),
                torch.linspace(
                    0, W_ - 1, W_, dtype=torch.float32, device=memory.device
                ),
            )
            grid = torch.cat([grid_x.unsqueeze(-1), grid_y.unsqueeze(-1)], -1)

            scale = torch.cat([valid_W.unsqueeze(-1), valid_H.unsqueeze(-1)], 1).view(
                N_, 1, 1, 2
            )
            grid = (grid.unsqueeze(0).expand(N_, -1, -1, -1) + 0.5) / scale
            wh = torch.ones_like(grid) * 0.05 * (2.0**lvl)
            proposal = torch.cat((grid, wh), -1).view(N_, -1, 4)
            proposals.append(proposal)
            _cur += H_ * W_
        output_proposals = torch.cat(proposals, 1)
        output_proposals_valid = (
            (output_proposals > 0.01) & (output_proposals < 0.99)
        ).all(-1, keepdim=True)
        output_proposals = torch.log(output_proposals / (1 - output_proposals))
        output_proposals = output_proposals.masked_fill(
            memory_padding_mask.unsqueeze(-1), float("inf")
        )
        output_proposals = output_proposals.masked_fill(
            ~output_proposals_valid, float("inf")
        )

        output_memory = memory
        output_memory = output_memory.masked_fill(
            memory_padding_mask.unsqueeze(-1), float(0)
        )
        output_memory = output_memory.masked_fill(~output_proposals_valid, float(0))
        output_memory = self.enc_output_norm(self.enc_output(output_memory))
        return output_memory, output_proposals

    def get_valid_ratio(self, mask):
        _, H, W = mask.shape
        valid_H = torch.sum(~mask[:, :, 0], 1)
        valid_W = torch.sum(~mask[:, 0, :], 1)
        valid_ratio_h = valid_H.float() / H
        valid_ratio_w = valid_W.float() / W
        valid_ratio = torch.stack([valid_ratio_w, valid_ratio_h], -1)
        return valid_ratio

    def forward(self, srcs, masks, pos_embeds, query_embed=None):
        assert self.two_stage or query_embed is not None

        # prepare input for encoder
        src_flatten = []
        mask_flatten = []
        lvl_pos_embed_flatten = []
        spatial_shapes = []
        for lvl, (src, mask, pos_embed) in enumerate(zip(srcs, masks, pos_embeds)):
            bs, c, h, w = src.shape
            spatial_shape = (h, w)
            spatial_shapes.append(spatial_shape)
            src = src.flatten(2).transpose(1, 2)
            mask = mask.flatten(1)
            pos_embed = pos_embed.flatten(2).transpose(1, 2)
            lvl_pos_embed = pos_embed + self.level_embed[lvl].view(1, 1, -1)
            lvl_pos_embed_flatten.append(lvl_pos_embed)
            src_flatten.append(src)
            mask_flatten.append(mask)
        src_flatten = torch.cat(src_flatten, 1)
        mask_flatten = torch.cat(mask_flatten, 1)
        lvl_pos_embed_flatten = torch.cat(lvl_pos_embed_flatten, 1)
        spatial_shapes = torch.as_tensor(
            spatial_shapes, dtype=torch.long, device=src_flatten.device
        )
        level_start_index = torch.cat(
            (spatial_shapes.new_zeros((1,)), spatial_shapes.prod(1).cumsum(0)[:-1])
        )
        valid_ratios = torch.stack([self.get_valid_ratio(m) for m in masks], 1)

        # encoder
        memory = self.encoder(
            src_flatten,
            spatial_shapes,
            level_start_index,
            valid_ratios,
            lvl_pos_embed_flatten,
            mask_flatten,
        )

        # prepare input for decoder
        bs, _, c = memory.shape
        if self.two_stage:
            output_memory, output_proposals = self.gen_encoder_output_proposals(
                memory, mask_flatten, spatial_shapes
            )

            # hack implementation for two-stage Deformable DETR
            enc_outputs_class = self.decoder.class_embed[self.decoder.num_layers](
                output_memory
            )
            enc_outputs_coord_unact = (
                self.decoder.bbox_embed[self.decoder.num_layers](output_memory)
                + output_proposals
            )

            topk = self.two_stage_num_proposals
            topk_proposals = torch.topk(enc_outputs_class[..., 0], topk, dim=1)[1]
            topk_coords_unact = torch.gather(
                enc_outputs_coord_unact, 1, topk_proposals.unsqueeze(-1).repeat(1, 1, 4)
            )
            topk_coords_unact = topk_coords_unact.detach()
            reference_points = topk_coords_unact.sigmoid()
            init_reference_out = reference_points
            pos_trans_out = self.pos_trans_norm(
                self.pos_trans(self.get_proposal_pos_embed(topk_coords_unact))
            )
            query_embed, tgt = torch.split(pos_trans_out, c, dim=2)
        else:
            query_embed, tgt = torch.split(query_embed, c, dim=1)
            query_embed = query_embed.unsqueeze(0).expand(bs, -1, -1)
            tgt = tgt.unsqueeze(0).expand(bs, -1, -1)
            reference_points = self.reference_points(query_embed).sigmoid()
            init_reference_out = reference_points

        # decoder
        hs, inter_references = self.decoder(
            tgt,
            reference_points,
            memory,
            spatial_shapes,
            level_start_index,
            valid_ratios,
            query_embed,
            mask_flatten,
        )

        inter_references_out = inter_references
        if self.two_stage:
            return (
                hs,
                init_reference_out,
                inter_references_out,
                enc_outputs_class,
                enc_outputs_coord_unact,
            )
        return hs, init_reference_out, inter_references_out, None, None


class DeformableTransformerEncoderLayer(nn.Module):
    def __init__(
        self,
        d_model=256,
        d_ffn=1024,
        dropout=0.1,
        activation="relu",
        n_levels=4,
        n_heads=8,
        n_points=4,
    ):
        super().__init__()

        # self attention
        self.self_attn = MSDeformAttn(d_model, n_levels, n_heads, n_points)
        self.dropout1 = nn.Dropout(dropout)
        self.norm1 = nn.LayerNorm(d_model)

        # ffn
        self.linear1 = nn.Linear(d_model, d_ffn)
        self.activation = _get_activation_fn(activation)
        self.dropout2 = nn.Dropout(dropout)
        self.linear2 = nn.Linear(d_ffn, d_model)
        self.dropout3 = nn.Dropout(dropout)
        self.norm2 = nn.LayerNorm(d_model)

    @staticmethod
    def with_pos_embed(tensor, pos):
        return tensor if pos is None else tensor + pos

    def forward_ffn(self, src):
        src2 = self.linear2(self.dropout2(self.activation(self.linear1(src))))
        src = src + self.dropout3(src2)
        src = self.norm2(src)
        return src

    def forward(
        self,
        src,
        pos,
        reference_points,
        spatial_shapes,
        level_start_index,
        padding_mask=None,
    ):
        # self attention
        src2 = self.self_attn(
            self.with_pos_embed(src, pos),
            reference_points,
            src,
            spatial_shapes,
            level_start_index,
            padding_mask,
        )
        src = src + self.dropout1(src2)
        src = self.norm1(src)

        # ffn
        src = self.forward_ffn(src)

        return src


class DeformableTransformerEncoder(nn.Module):
    def __init__(self, encoder_layer, num_layers):
        super().__init__()
        self.layers = _get_clones(encoder_layer, num_layers)
        self.num_layers = num_layers

    @staticmethod
    def get_reference_points(spatial_shapes, valid_ratios, device):
        reference_points_list = []
        for lvl, (H_, W_) in enumerate(spatial_shapes):
            ref_y, ref_x = torch.meshgrid(
                torch.linspace(0.5, H_ - 0.5, H_, dtype=torch.float32, device=device),
                torch.linspace(0.5, W_ - 0.5, W_, dtype=torch.float32, device=device),
            )
            ref_y = ref_y.reshape(-1)[None] / (valid_ratios[:, None, lvl, 1] * H_)
            ref_x = ref_x.reshape(-1)[None] / (valid_ratios[:, None, lvl, 0] * W_)
            ref = torch.stack((ref_x, ref_y), -1)
            reference_points_list.append(ref)
        reference_points = torch.cat(reference_points_list, 1)
        reference_points = reference_points[:, :, None] * valid_ratios[:, None]
        return reference_points

    def forward(
        self,
        src,
        spatial_shapes,
        level_start_index,
        valid_ratios,
        pos=None,
        padding_mask=None,
    ):
        output = src
        reference_points = self.get_reference_points(
            spatial_shapes, valid_ratios, device=src.device
        )
        for _, layer in enumerate(self.layers):
            output = layer(
                output,
                pos,
                reference_points,
                spatial_shapes,
                level_start_index,
                padding_mask,
            )

        return output


class DeformableTransformerDecoderLayer(nn.Module):
    def __init__(
        self,
        d_model=256,
        d_ffn=1024,
        dropout=0.1,
        activation="relu",
        n_levels=4,
        n_heads=8,
        n_points=4,
    ):
        super().__init__()

        # cross attention
        self.cross_attn = MSDeformAttn(d_model, n_levels, n_heads, n_points)
        self.dropout1 = nn.Dropout(dropout)
        self.norm1 = nn.LayerNorm(d_model)

        # self attention
        self.self_attn = nn.MultiheadAttention(d_model, n_heads, dropout=dropout)
        self.dropout2 = nn.Dropout(dropout)
        self.norm2 = nn.LayerNorm(d_model)

        # ffn
        self.linear1 = nn.Linear(d_model, d_ffn)
        self.activation = _get_activation_fn(activation)
        self.dropout3 = nn.Dropout(dropout)
        self.linear2 = nn.Linear(d_ffn, d_model)
        self.dropout4 = nn.Dropout(dropout)
        self.norm3 = nn.LayerNorm(d_model)

    @staticmethod
    def with_pos_embed(tensor, pos):
        return tensor if pos is None else tensor + pos

    def forward_ffn(self, tgt):
        tgt2 = self.linear2(self.dropout3(self.activation(self.linear1(tgt))))
        tgt = tgt + self.dropout4(tgt2)
        tgt = self.norm3(tgt)
        return tgt

    def forward(
        self,
        tgt,
        query_pos,
        reference_points,
        src,
        src_spatial_shapes,
        level_start_index,
        src_padding_mask=None,
    ):
        # self attention
        q = k = self.with_pos_embed(tgt, query_pos)
        tgt2 = self.self_attn(
            q.transpose(0, 1), k.transpose(0, 1), tgt.transpose(0, 1)
        )[0].transpose(0, 1)
        tgt = tgt + self.dropout2(tgt2)
        tgt = self.norm2(tgt)

        # cross attention
        tgt2 = self.cross_attn(
            self.with_pos_embed(tgt, query_pos),
            reference_points,
            src,
            src_spatial_shapes,
            level_start_index,
            src_padding_mask,
        )
        tgt = tgt + self.dropout1(tgt2)
        tgt = self.norm1(tgt)

        # ffn
        tgt = self.forward_ffn(tgt)

        return tgt


class DeformableTransformerDecoder(nn.Module):
    def __init__(self, decoder_layer, num_layers, return_intermediate=False):
        super().__init__()
        self.layers = _get_clones(decoder_layer, num_layers)
        self.num_layers = num_layers
        self.return_intermediate = return_intermediate
        # hack implementation for iterative bounding box refinement and two-stage Deformable DETR
        self.bbox_embed = None
        self.class_embed = None

    def forward(
        self,
        tgt,
        reference_points,
        src,
        src_spatial_shapes,
        src_level_start_index,
        src_valid_ratios,
        query_pos=None,
        src_padding_mask=None,
    ):
        output = tgt

        intermediate = []
        intermediate_reference_points = []
        for lid, layer in enumerate(self.layers):
            if reference_points.shape[-1] == 4:
                reference_points_input = (
                    reference_points[:, :, None]
                    * torch.cat([src_valid_ratios, src_valid_ratios], -1)[:, None]
                )
            else:
                assert reference_points.shape[-1] == 2
                reference_points_input = (
                    reference_points[:, :, None] * src_valid_ratios[:, None]
                )
            output = layer(
                output,
                query_pos,
                reference_points_input,
                src,
                src_spatial_shapes,
                src_level_start_index,
                src_padding_mask,
            )

            # hack implementation for iterative bounding box refinement
            if self.bbox_embed is not None:
                tmp = self.bbox_embed[lid](output)
                if reference_points.shape[-1] == 4:
                    new_reference_points = tmp + inverse_sigmoid(reference_points)
                    new_reference_points = new_reference_points.sigmoid()
                else:
                    assert reference_points.shape[-1] == 2
                    new_reference_points = tmp
                    new_reference_points[..., :2] = tmp[..., :2] + inverse_sigmoid(
                        reference_points
                    )
                    new_reference_points = new_reference_points.sigmoid()
                reference_points = new_reference_points.detach()

            if self.return_intermediate:
                intermediate.append(output)
                intermediate_reference_points.append(reference_points)

        if self.return_intermediate:
            return torch.stack(intermediate), torch.stack(intermediate_reference_points)

        return output, reference_points


def _get_clones(module, N):
    return nn.ModuleList([copy.deepcopy(module) for i in range(N)])


def _get_activation_fn(activation):
    """Return an activation function given a string"""
    if activation == "relu":
        return F.relu
    if activation == "gelu":
        return F.gelu
    if activation == "glu":
        return F.glu
    raise RuntimeError(f"activation should be relu/gelu, not {activation}.")


def build_deforamble_transformer(args):
    return DeformableTransformer(
        d_model=args.hidden_dim,
        nhead=args.nheads,
        num_encoder_layers=args.enc_layers,
        num_decoder_layers=args.dec_layers,
        dim_feedforward=args.dim_feedforward,
        dropout=args.dropout,
        activation="relu",
        return_intermediate_dec=True,
        num_feature_levels=args.num_feature_levels,
        dec_n_points=args.dec_n_points,
        enc_n_points=args.enc_n_points,
        two_stage=args.two_stage,
        two_stage_num_proposals=args.num_queries,
    )


def build_learn_deforamble_transformer():
    return DeformableTransformer(
        d_model=256,
        nhead=8,
        num_encoder_layers=6,
        num_decoder_layers=6,
        dim_feedforward=1024,
        dropout=0.1,
        activation="relu",
        return_intermediate_dec=True,
        num_feature_levels=4,
        dec_n_points=4,
        enc_n_points=4,
        two_stage=False,
        two_stage_num_proposals=100,
    )



import re
import os
import sys
import copy
from copy import deepcopy
import types
import random
import warnings
import traceback
import pandas as pd
from functools import partial

HAS_FASTAI = True
try:
    from enum import Enum
    import torch
    import arcgis
    from fastprogress.fastprogress import progress_bar
    from fastai.text import (
        TextList,
        TextClasDataBunch,
        pad_collate,
        TextDataBunch,
        SortishSampler,
        SortSampler,
        ItemList,
        ItemBase,
        Text,
    )
    from fastai.text import *
    from fastai.layers import CrossEntropyFlat
    from fastai.basic_train import Learner
    from fastai.torch_core import to_data
    from fastai.callback import Callback
except Exception as e:
    import_exception = "\n".join(
        traceback.format_exception(type(e), e, e.__traceback__)
    )
    HAS_FASTAI = False

HAS_NUMPY = True
try:
    import numpy as np

    warnings.filterwarnings("ignore", category=np.VisibleDeprecationWarning)
except:
    HAS_NUMPY = False


max_len = 100
DatasetType = Enum("DatasetType", "Train Valid Test Single Fix")


class SequenceToSequenceDataBunch(TextDataBunch):
    "Create a `TextDataBunch` suitable for training."

    @classmethod
    def create(
        cls,
        train_ds,
        valid_ds,
        test_ds=None,
        path=".",
        bs=32,
        pad_idx=1,
        dl_tfms=None,
        pad_first=False,
        no_check=False,
        backwards=False,
        val_bs=None,
        collate_fn=None,
        **dl_kwargs
    ):
        "Function that transform the `datasets` in a `DataBunch`. Passes `**dl_kwargs` on to `DataLoader()`"
        device = dl_kwargs.pop("device", None)
        from torch.utils.data import DataLoader

        datasets = cls._init_ds(train_ds, valid_ds)
        val_bs = bs
        collate_fn = partial(
            seq2seq_collate, pad_idx=pad_idx, pad_first=pad_first, backwards=backwards
        )
        train_sampler = SortishSampler(
            datasets[0].x, key=lambda t: len(datasets[0][t][0].data), bs=bs // 2
        )
        train_dl = DataLoader(
            datasets[0],
            batch_size=bs,
            sampler=train_sampler,
            drop_last=True,
            **dl_kwargs
        )
        dataloaders = [train_dl]
        for i, ds in enumerate(datasets[1:]):
            lengths = [len(t) for t in ds.x.items]
            sampler = SortSampler(ds.x, key=lengths.__getitem__)
            dataloaders.append(
                DataLoader(ds, batch_size=val_bs, sampler=sampler, **dl_kwargs)
            )
        return cls(
            *dataloaders,
            dl_tfms=dl_tfms,
            path=path,
            collate_fn=collate_fn,
            no_check=no_check,
            device=device
        )


class SequenceToSequenceTextList(TextList):
    _bunch = SequenceToSequenceDataBunch
    _label_cls = TextList

    def reconstruct(self, t: torch.Tensor):
        if isinstance(t, (list)):  # handling the 'shift_tfm' dataloader transform
            t = t[0]
        elif t.ndim == 2:
            t = t.argmax(-1)
        idx_min = (t != self.pad_idx).nonzero().min()
        idx_max = (t != self.pad_idx).nonzero().max()
        return Text(
            t[idx_min : idx_max + 1], self.vocab.textify(t[idx_min : idx_max + 1])
        )


class SequenceToSequenceLearner(Learner):
    def __init__(self, data, model, metrics=None, loss_func=None, **learn_kwargs):
        # self.model_name = model_name
        pretrained_model_name = model.pretrained_model_name
        super().__init__(
            data, model, metrics=metrics, loss_func=loss_func, **learn_kwargs
        )
        self.tokenizer = (
            self.model._tokenizer
        )  # AutoTokenizer.from_pretrained(pretrained_model_name)
        if self.tokenizer.__class__.__name__ == "MarianTokenizer":
            self.tokenizer.clean_up_tokenization = clean_up_tokenization

    def show_results(
        self, ds_type=DatasetType.Valid, rows: int = 5, num_beams=4, max_len=50
    ):
        if not HAS_NUMPY:
            raise Exception("This function requires numpy.")
        from IPython.display import display, HTML

        "Show `rows` result of predictions on `ds_type` dataset."
        ds = self.dl(ds_type).dataset
        x, y = self.data.one_batch(ds_type, detach=False, denorm=False)
        preds = self.pred_batch(batch=(x, y))
        xs = [ds.x.reconstruct(grab_idx(x, i)) for i in range(rows)]
        ys = [ds.x.reconstruct(grab_idx(y, i)) for i in range(rows)]
        zs = self.model._transformer.generate(
            x[0], num_beams=num_beams, max_length=max_len
        )
        items, names = [], ["text", "target", "pred"]
        for i, (x, y, z) in enumerate(zip(xs, ys, zs)):
            txt_x = self.tokenizer.decode(x.data, skip_special_tokens=True)
            txt_y = self.tokenizer.decode(y.data, skip_special_tokens=True)
            txt_z = self.tokenizer.decode(z, skip_special_tokens=True)
            items.append([txt_x, txt_y, txt_z])
        items = np.array(items)
        df = pd.DataFrame({n: items[:, i] for i, n in enumerate(names)}, columns=names)
        with pd.option_context("display.max_colwidth", 0):
            display(HTML(df.to_html(index=False)))

    def predict_batch(self, batch_text, num_beams, max_length, min_length):
        tok = self.model._tokenizer
        encoded_input_batch = tok.batch_encode_plus(
            batch_text, padding=True, return_tensors="pt"
        )["input_ids"]
        encoded_output_batch = self.model._transformer.generate(
            encoded_input_batch.to(self.model._transformer.device.type),
            num_beams=num_beams,
            max_length=max_length,
            min_length=min_length,
        )
        decoded_output_batch = tok.batch_decode(
            encoded_output_batch, skip_special_tokens=True
        )
        return decoded_output_batch, encoded_output_batch

    def predict(
        self, text_list, batch_size, show_progress, **kwargs
    ):  # num_beams=4, max_len=50):
        tok = self.model._tokenizer
        num_beams = kwargs.get("num_beams", 1)
        max_length = kwargs.get("max_length", 20)
        min_length = kwargs.get("min_length", 10)

        if isinstance(text_list, (list)):
            decoded_output = []
            for i in progress_bar(
                range(0, len(text_list), batch_size), display=show_progress
            ):
                batch_text = text_list[i : i + batch_size]
                decoded_output_batch, encoded_output_batch = self.predict_batch(
                    batch_text,
                    num_beams=num_beams,
                    max_length=max_length,
                    min_length=min_length,
                )
                decoded_output.extend(decoded_output_batch)
        else:
            encoded_input = text_list
            # encoded_input.unsqueeze_(0)
            encoded_output = self.model._transformer.generate(
                encoded_input.to(self.model._transformer.device.type),
                num_beams=num_beams,
                max_length=max_length,
                min_length=min_length,
            )
            decoded_output = tok.batch_decode(encoded_output, skip_special_tokens=True)
        return decoded_output


class NGram:
    if not HAS_NUMPY:
        raise Exception("This function requires numpy.")

    def __init__(self, ngram, max_n=5000):
        self.ngram, self.max_n = ngram, max_n

    def __eq__(self, other):
        if len(self.ngram) != len(other.ngram):
            return False
        return np.all(np.array(self.ngram) == np.array(other.ngram))

    def __hash__(self):
        return int(sum([o * self.max_n**i for i, o in enumerate(self.ngram)]))


def get_grams(x, n, max_n=5000):
    return (
        x
        if n == 1
        else [NGram(x[i : i + n], max_n=max_n) for i in range(len(x) - n + 1)]
    )


def get_correct_ngrams(pred, targ, n, max_n=5000):
    pred_grams, targ_grams = (
        get_grams(pred, n, max_n=max_n),
        get_grams(targ, n, max_n=max_n),
    )
    pred_cnt, targ_cnt = Counter(pred_grams), Counter(targ_grams)
    return sum([min(c, targ_cnt[g]) for g, c in pred_cnt.items()]), len(pred_grams)


class CorpusBLEU(Callback):
    def __init__(self, vocab_sz):
        self.vocab_sz = vocab_sz
        self.name = "bleu"

    def on_epoch_begin(self, **kwargs):
        self.pred_len, self.targ_len, self.corrects, self.counts = (
            0,
            0,
            [0] * 4,
            [0] * 4,
        )

    def on_batch_end(self, last_output, last_target, **kwargs):
        loss, last_output = last_output
        last_output = last_output.argmax(dim=-1)
        for pred, targ in zip(last_output.cpu().numpy(), last_target.cpu().numpy()):
            self.pred_len += len(pred)
            self.targ_len += len(targ)
            for i in range(4):
                c, t = get_correct_ngrams(pred, targ, i + 1, max_n=self.vocab_sz)
                self.corrects[i] += c
                self.counts[i] += t

    def on_epoch_end(self, last_metrics, **kwargs):
        precs = [c / t for c, t in zip(self.corrects, self.counts)]
        len_penalty = (
            exp(1 - self.targ_len / self.pred_len)
            if self.pred_len < self.targ_len
            else 1
        )
        bleu = len_penalty * ((precs[0] * precs[1] * precs[2] * precs[3]) ** 0.25)
        return add_metrics(last_metrics, bleu)


def seq2seq_loss(out, targ, pad_idx=1):
    loss, _ = out
    return loss


def seq2seq_acc(out, targ, pad_idx=1):
    loss, out = out
    bs, targ_len = targ.size()
    _, out_len, vs = out.size()
    if targ_len > out_len:
        out = F.pad(out, (0, 0, 0, targ_len - out_len, 0, 0), value=pad_idx)
    if out_len > targ_len:
        targ = F.pad(targ, (0, out_len - targ_len, 0, 0), value=pad_idx)
    out = out.argmax(2)
    return (out == targ).float().mean()


def seq2seq_collate(samples, pad_idx=1, pad_first=True, backwards=False):
    "Function that collect samples and adds padding. Flips token order if needed"
    samples = to_data(samples)
    max_len_x, max_len_y = (
        max([len(s[0]) for s in samples]),
        max([len(s[1]) for s in samples]),
    )
    res_x = torch.zeros(len(samples), max_len_x).long() + pad_idx
    res_y = torch.zeros(len(samples), max_len_y).long() + pad_idx
    if backwards:
        pad_first = not pad_first
    for i, s in enumerate(samples):
        if pad_first:
            res_x[i, -len(s[0]) :], res_y[i, -len(s[1]) :] = (
                torch.LongTensor(s[0]),
                torch.LongTensor(s[1]),
            )
        else:
            res_x[i, : len(s[0])], res_y[i, : len(s[1])] = (
                torch.LongTensor(s[0]),
                torch.LongTensor(s[1]),
            )
    if backwards:
        res_x, res_y = res_x.flip(1), res_y.flip(1)
    return res_x, res_y


# def shift_tfm(b):
#     #transform to modify input for teacher forcing
#     x,y = b
#     y = torch.nn.functional.pad(y, (1, 0), value=0)
#     return [x,y[:,:-1]], y[:,1:]


def teacher_forcing_tfm(b):
    x, y = b
    return [x, y], y


def clean_up_tokenization(out_string: str) -> str:
    """
    Additional handling of '▁' in clean_up_tokenization for marian tokenizer.
    """
    out_string = (
        out_string.replace(" .", ".")
        .replace(" ?", "?")
        .replace(" !", "!")
        .replace(" ,", ",")
        .replace(" ' ", "'")
        .replace(" n't", "n't")
        .replace(" 'm", "'m")
        .replace(" 's", "'s")
        .replace(" 've", "'ve")
        .replace(" 're", "'re")
    )
    out_string = re.sub(r"▁(?=\S)", " ", out_string)
    return out_string



import os
import glob
import uuid
import importlib
from functools import lru_cache
import numpy as np
import pandas as pd
from distutils.version import LooseVersion
import dask
from dask.dataframe import Series, from_pandas
import dask.dataframe as dd
from dask.dataframe.core import get_parallel_type, make_meta
from dask.base import normalize_token
from dask.dataframe.extensions import (
    make_array_nonempty,
    make_scalar,
    register_dataframe_accessor,
    register_series_accessor,
)

DASK_2021_06_0 = str(dask.__version__) >= LooseVersion("2021.06.0")

if DASK_2021_06_0:
    from dask.dataframe.dispatch import make_meta_dispatch
    from dask.dataframe.backends import (
        _nonempty_index,
        meta_nonempty_dataframe,
        meta_nonempty,
    )
else:
    from dask.dataframe.core import make_meta as make_meta_dispatch
    from dask.dataframe.utils import (
        _nonempty_index,
        meta_nonempty_dataframe,
        meta_nonempty,
    )

from arcgis.geometry import BaseGeometry, Geometry
from arcgis.features import FeatureCollection
from arcgis._impl.common._isd import InsensitiveDict
from arcgis._impl.common._mixins import PropertyMap
from ._array import GeoType, GeoArray
from ._io.fileops import to_featureclass, from_featureclass
from ._index._dqtree import DaskSpatialIndex
from ._viz._dmapping import dask_plot


# -------------------------------------------------------------------------
@lru_cache(maxsize=200)
def _default_geometry():
    return Geometry({"x": 0, "y": 0, "spatialReference": {"wkid": 4326}})


# -------------------------------------------------------------------------
def _isna(value):
    """
    Check if scalar value is NA-like (None or np.nan).
    Custom version that only works for scalars (returning True or False),
    as `pd.isna` also works for array-like input returning a boolean array.
    """
    if value is None:
        return True
    elif isinstance(value, float) and np.isnan(value):
        return True
    else:
        return False
    # -------------------------------------------------------------------------


def _from_geometry(data: list) -> "GeoArray":
    """
    Convert a list or array of dict/geometries objects to a GeoArray.
    """

    n = len(data)

    out = []

    for idx in range(n):
        geom = data[idx]
        if isinstance(geom, BaseGeometry):
            out.append(geom)
        elif hasattr(geom, "__geo_interface__"):
            geom = Geometry(geom.__geo_interface__)
            out.append(geom)
        elif _isna(geom):
            out.append(None)
        else:
            raise TypeError("Input must be valid geometry objects: {0}".format(geom))

    aout = np.empty(n, dtype=object)
    aout[:] = out
    return GeoArray(aout)


# -------------------------------------------------------------------------
@register_dataframe_accessor("spatial")
class GeoDaskSpatialAccessor:
    """
    The Geospatial dask dataframe accessor used to perform dataset specific operations.
    """

    _sr = None
    _data = None
    _name = None
    _sindex = None
    _renderer = None

    # ----------------------------------------------------------------------
    def __init__(self, obj):
        self._data = obj
        self._fn_attr = lambda a, op: getattr(a, op)
        self._fn_method = lambda a, op, **kwargs: getattr(a, op)(**kwargs)
        self._sfn_method = lambda a, sop, **kwargs: getattr(a, sop)(**kwargs)

    # ----------------------------------------------------------------------
    @property
    def __feature_set__(self):
        """returns a dictionary representation of an Esri FeatureSet"""
        row_pieces = self._data.map_partitions(
            lambda part: part.spatial.__feature_set__
        ).compute()

        if len(row_pieces) == 1:
            return row_pieces[0]
        else:
            fs = row_pieces[0]
            [fs["features"].extend(pieces["features"]) for pieces in row_pieces[1:]]
            return fs

    # ----------------------------------------------------------------------
    @dask.delayed
    def to_feature_collection(
        self, name=None, drawing_info=None, extent=None, global_id_field=None
    ):
        """
        Converts a Spatially Enabled Dask DataFrame to a Feature Collection

        =====================  ===============================================================
        **optional argument**  **Description**
        ---------------------  ---------------------------------------------------------------
        name                   optional string. Name of the Feature Collection
        ---------------------  ---------------------------------------------------------------
        drawing_info           Optional dictionary. This is the rendering information for a
                               Feature Collection.  Rendering information is a dictionary with
                               the symbology, labelling and other properties defined.  See:
                               https://developers.arcgis.com/documentation/common-data-types/renderer-objects.htm
        ---------------------  ---------------------------------------------------------------
        extent                 Optional dictionary.  If desired, a custom extent can be
                               provided to set where the map starts up when showing the data.
                               The default is the full extent of the dataset in the Spatial
                               DataFrame.
        ---------------------  ---------------------------------------------------------------
        global_id_field        Optional string. The Global ID field of the dataset.
        =====================  ===============================================================

        :returns: FeatureCollection object
        """

        import string
        import random

        fs = self.__feature_set__
        if name is None:
            name = random.choice(string.ascii_letters) + uuid.uuid4().hex[:5]
        if extent is None:
            ext = self.full_extent
            extent = {
                "xmin": ext[0],
                "ymin": ext[1],
                "xmax": ext[2],
                "ymax": ext[3],
                "spatialReference": fs["spatialReference"],
            }

        for fld in fs["fields"]:
            if fld["name"].lower() == fs["objectIdFieldName"].lower():
                fld["editable"] = False
                fld["sqlType"] = "sqlTypeOther"
                fld["domain"] = None
                fld["defaultValue"] = None
                fld["nullable"] = False
            else:
                fld["editable"] = True
                fld["sqlType"] = "sqlTypeOther"
                fld["domain"] = None
                fld["defaultValue"] = None
                fld["nullable"] = True
        if drawing_info is None:
            import json

            di = {"renderer": json.loads(self._data.spatial.renderer.json)}
        else:
            di = drawing_info
        layer = {
            "layerDefinition": {
                "currentVersion": 10.7,
                "id": 0,
                "name": name,
                "type": "Feature Layer",
                "displayField": "",
                "description": "",
                "copyrightText": "",
                "defaultVisibility": True,
                "relationships": [],
                "isDataVersioned": False,
                "supportsAppend": True,
                "supportsCalculate": True,
                "supportsASyncCalculate": True,
                "supportsTruncate": False,
                "supportsAttachmentsByUploadId": True,
                "supportsAttachmentsResizing": True,
                "supportsRollbackOnFailureParameter": True,
                "supportsStatistics": True,
                "supportsExceedsLimitStatistics": True,
                "supportsAdvancedQueries": True,
                "supportsValidateSql": True,
                "supportsCoordinatesQuantization": True,
                "supportsFieldDescriptionProperty": True,
                "supportsQuantizationEditMode": True,
                "supportsApplyEditsWithGlobalIds": False,
                "supportsMultiScaleGeometry": True,
                "supportsReturningQueryGeometry": True,
                "hasGeometryProperties": True,
                "advancedQueryCapabilities": {
                    "supportsPagination": True,
                    "supportsPaginationOnAggregatedQueries": True,
                    "supportsQueryRelatedPagination": True,
                    "supportsQueryWithDistance": True,
                    "supportsReturningQueryExtent": True,
                    "supportsStatistics": True,
                    "supportsOrderBy": True,
                    "supportsDistinct": True,
                    "supportsQueryWithResultType": True,
                    "supportsSqlExpression": True,
                    "supportsAdvancedQueryRelated": True,
                    "supportsCountDistinct": True,
                    "supportsReturningGeometryCentroid": True,
                    "supportsReturningGeometryProperties": True,
                    "supportsQueryWithDatumTransformation": True,
                    "supportsHavingClause": True,
                    "supportsOutFieldSQLExpression": True,
                    "supportsMaxRecordCountFactor": True,
                    "supportsTopFeaturesQuery": True,
                    "supportsDisjointSpatialRel": True,
                    "supportsQueryWithCacheHint": True,
                },
                "useStandardizedQueries": False,
                "geometryType": fs["geometryType"],
                "minScale": 0,
                "maxScale": 0,
                "extent": extent,
                "drawingInfo": di,
                "allowGeometryUpdates": True,
                "hasAttachments": False,
                "htmlPopupType": "esriServerHTMLPopupTypeNone",
                "hasM": False,
                "hasZ": False,
                "objectIdField": fs["objectIdFieldName"] or "OBJECTID",
                "globalIdField": "",
                "typeIdField": "",
                "fields": fs["fields"],
                "types": [],
                "supportedQueryFormats": "JSON, geoJSON",
                "hasStaticData": True,
                "maxRecordCount": 32000,
                "standardMaxRecordCount": 4000,
                "tileMaxRecordCount": 4000,
                "maxRecordCountFactor": 1,
                "capabilities": "Query",
            },
            "featureSet": {
                "features": fs["features"],
                "geometryType": fs["geometryType"],
            },
        }
        if global_id_field is not None:
            layer["layerDefinition"]["globalIdField"] = global_id_field
        return FeatureCollection(layer)

    # ----------------------------------------------------------------------
    @property
    def name(self):
        """returns the name of the geometry column"""
        if self._name is None:
            try:
                cols = [c.lower() for c in self._data.columns]
                if any(self._data.dtypes == "geometry"):
                    name = self._data.dtypes[self._data.dtypes == "geometry"].index[0]
                    self._name = name
                    return name
                elif "shape" in cols:
                    idx = cols.index("shape")
                    self._name = self._data.columns[idx]
            except:
                raise Exception("Spatial column not defined, please use `set_geometry`")
        return self._name

    # ----------------------------------------------------------------------
    @property
    def area(self):
        """
        Returns the total area of the dataframe

        :returns: float

        >>> df.spatial.area
        143.23427

        """
        return self._data[self.name].geom.area.sum()

    # ----------------------------------------------------------------------
    @property
    def bbox(self):
        """
        Returns the total length of the dataframe

        :returns: Polygon

        >>> df.spatial.bbox
        {'rings' : [[[1,2], [2,3], [3,3],....]], 'spatialReference' {'wkid': 4326}}
        """
        parts = self._data.map_partitions(
            lambda part: self._fn_attr(part.spatial, "bbox")
        ).compute()
        if len(parts) == 1:
            df = pd.DataFrame(parts[0]["rings"][0], columns=["x", "y"])
            sr = parts[0]["spatialReference"]
        else:
            main_part = parts[0]
            sr = parts[0]["spatialReference"]
            for pt in parts[1:]:
                main_part["rings"][0].extend(pt["rings"][0])
            df = pd.DataFrame(main_part["rings"][0], columns=["x", "y"])
        xmin, ymin, xmax, ymax = (
            df.x.min(),
            df.y.min(),
            df.x.max(),
            df.y.max(),
        )
        if isinstance(sr, list) and len(sr) > 0:
            sr = sr[0]
        if xmin == xmax:
            xmin -= 0.001
            xmax += 0.001
        if ymin == ymax:
            ymin -= 0.001
            ymax += 0.001
        return Geometry(
            {
                "rings": [
                    [
                        [xmin, ymin],
                        [xmin, ymax],
                        [xmax, ymax],
                        [xmax, ymin],
                        [xmin, ymin],
                    ]
                ],
                "spatialReference": dict(sr),
            }
        )

    # ----------------------------------------------------------------------
    @property
    def centroid(self):
        """
        Returns the centroid of the dataframe

        :returns: Geometry

        >>> df.spatial.centroid
        (-14.23427, 39)

        """

        df = pd.DataFrame(
            self._data[self.name].geom.centroid.compute().tolist(),
            columns=["x", "y"],
        )
        return df.x.mean(), df.y.mean()

    # ----------------------------------------------------------------------
    @property
    def full_extent(self):
        """
        Returns the extent of the dataframe

        :returns: tuple

        >>> df.spatial.full_extent
        (-118, 32, -97, 33)

        """
        ge = self._data[self.name].geom.extent.compute()
        q = ge.notnull()
        data = ge[q].tolist()
        array = np.array(data)
        return (
            float(array[:, 0][array[:, 0] != None].min()),
            float(array[:, 1][array[:, 1] != None].min()),
            float(array[:, 2][array[:, 2] != None].max()),
            float(array[:, 3][array[:, 3] != None].max()),
        )

    # ----------------------------------------------------------------------
    @property
    @lru_cache(maxsize=10)
    def geometry_type(self):
        """
        Returns a list Geometry Types for the DataFrame
        """
        q = self._data.index < 10
        return (
            self._data.loc[q, [self.name]][  # source dataframe
                self.name
            ]  # gets the SHAPE column and records
            .geom.geometry_type.unique()  # geometry accessor  # method to call  # get unique values
            .compute()  # run the command
            .tolist()  # convert to list
        )  # return results

    # ----------------------------------------------------------------------
    def relationship(self, other, op, relation=None):
        """
        This method allows for dataframe to dataframe compairson using
        spatial relationships.  The return is a pd.DataFrame that meet the
        operations' requirements.

        =========================    =========================================================
        **Parameter**                 **Description**
        -------------------------    ---------------------------------------------------------
        other                        Required Spatially Enabled DataFrame. The geometry to
                                     perform the operation from.
        -------------------------    ---------------------------------------------------------
        op                           Optional String. The spatial operation to perform.  The
                                     allowed value are: contains,crosses,disjoint,equals,
                                     overlaps,touches, or within.

                                     - contains - Indicates if the base geometry contains the comparison geometry.
                                     - crosses -  Indicates if the two geometries intersect in a geometry of a lesser shape type.
                                     - disjoint - Indicates if the base and comparison geometries share no points in common.
                                     - equals - Indicates if the base and comparison geometries are of the same shape type and define the same set of points in the plane. This is a 2D comparison only; M and Z values are ignored.
                                     - overlaps - Indicates if the intersection of the two geometries has the same shape type as one of the input geometries and is not equivalent to either of the input geometries.
                                     - touches - Indicates if the boundaries of the geometries intersect.
                                     - within - Indicates if the base geometry is within the comparison geometry.
                                     - intersect - Intdicates if the base geometry has an intersection of the other geometry.
        -------------------------    ---------------------------------------------------------
        relation                     Optional String.  The spatial relationship type.  The
                                     allowed values are: BOUNDARY, CLEMENTINI, and PROPER.

                                     + BOUNDARY - Relationship has no restrictions for interiors or boundaries.
                                     + CLEMENTINI - Interiors of geometries must intersect. This is the default.
                                     + PROPER - Boundaries of geometries must not intersect.

                                     This only applies to contains,
        =========================    =========================================================

        :returns: Delayed DataFrame


        """
        results = self._data.map_partitions(
            lambda part: self._fn_method(
                part.spatial,
                "relationship",
                **{"other": other, "relation": relation, "op": op},
            )
        ).compute()
        return results

    # ----------------------------------------------------------------------
    def select(self, other: "Geometry"):
        """
        This operation performs a dataset wide **selection** by geometric
        intersection. A geometry or another Spatially enabled DataFrame
        can be given and `select` will return all rows that intersect that
        input geometry.  The `select` operation uses a spatial index to
        complete the task, so if it is not built before the first run, the
        function will build a quadtree index on the fly.

        **requires ArcPy or Shapely**

        :returns: pd.DataFrame (spatially enabled)

        """
        results = self._data.map_partitions(
            lambda part: self._fn_method(part.spatial, "select", **{"other": other})
        )
        return results

    # ----------------------------------------------------------------------
    def overlay(self, sdf, op="union"):
        """
        Performs spatial operation operations on two spatially enabled dataframes.

        **requires ArcPy or Shapely**

        =========================    =========================================================
        **Parameter**                 **Description**
        -------------------------    ---------------------------------------------------------
        sdf                          Required Spatially Enabled DataFrame. The geometry to
                                     perform the operation from.
        -------------------------    ---------------------------------------------------------
        op                           Optional String. The spatial operation to perform.  The
                                     allowed value are: union, erase, identity, intersection.
                                     `union` is the default operation.
        =========================    =========================================================

        :returns: Spatially enabled DataFrame (pd.DataFrame)

        """
        from ._tools._dask_overlay import overlay_dask

        if isinstance(sdf, pd.DataFrame):
            meta = pd.concat([sdf[0:0], self._data._meta])

        elif isinstance(sdf, dd.DataFrame):
            meta = pd.concat([sdf._meta, self._data._meta])

        # joined_meta = pd.concat([ddf2._meta, ddf1._meta])
        op = str(op).lower()
        if self._data.spatial.geometry_type != sdf.spatial.geometry_type:
            raise ValueError(
                ("Spatially enabled DataFrame must " "be the same geometry type.")
            )

        if (
            op == "symmetric_difference"
            and self._data.spatial.geometry_type != ["polygon"]
            and sdf.spatial.geometry_type != ["polygon"]
        ):
            raise ValueError(
                ("symmetric_difference is only supported for " "polygon geometries.")
            )

        def fn(part):
            return overlay_dask(part, sdf, op)

        results = self._data.map_partitions(fn, meta=meta, enforce_metadata=False)
        return results

    # ----------------------------------------------------------------------
    def join(
        self,
        right_df,
        how="inner",
        op="intersects",
        left_tag="left",
        right_tag="right",
    ):
        """
        Joins the current DataFrame to another spatially enabled dataframes based
        on spatial location based.

        .. note::
            requires the SEDF to be in the same coordinate system


        ======================    =========================================================
        **Parameter**              **Description**
        ----------------------    ---------------------------------------------------------
        right_df                  Required pd.DataFrame. Spatially enabled dataframe to join.
        ----------------------    ---------------------------------------------------------
        how                       Required string. The type of join:

                                    + `left` - use keys from current dataframe and retains only current geometry column
                                    + `right` - use keys from right_df; retain only right_df geometry column
                                    + `inner` - use intersection of keys from both dfs and retain only current geometry column

        ----------------------    ---------------------------------------------------------
        op                        Required string. The operation to use to perform the join.
                                  The default is `intersects`.

                                  supported perations: `intersects`, `within`, and `contains`
        ----------------------    ---------------------------------------------------------
        left_tag                  Optional String. If the same column is in the left and
                                  right dataframe, this will append that string value to
                                  the field.
        ----------------------    ---------------------------------------------------------
        right_tag                 Optional String. If the same column is in the left and
                                  right dataframe, this will append that string value to
                                  the field.
        ======================    =========================================================

        :returns:
          Delayed Dask DataFrame
        """
        if isinstance(right_df, pd.DataFrame):
            meta = pd.concat([right_df[0:0], self._data._meta])

        elif isinstance(right_df, dd.DataFrame):
            meta = pd.concat([right_df._meta, self._data._meta])

        results = self._data.map_partitions(
            lambda part: self._sfn_method(
                part.spatial,
                "join",
                **{
                    "right_df": right_df,
                    "how": how,
                    "op": op,
                    "left_tag": left_tag,
                    "right_tag": right_tag,
                },
            ),
            meta=meta,
            enforce_metadata=False,
        )
        return results

    # ----------------------------------------------------------------------
    @property
    def has_z(self):
        """
        Returns a boolean that determines if the datasets have `Z` values

        :returns: Boolean
        """
        return self._data[self.name].geom.has_z.all()

    # ----------------------------------------------------------------------
    @property
    def has_m(self):
        """
        Returns a boolean that determines if the datasets have `Z` values

        :returns: Boolean
        """
        return self._data[self.name].geom.has_m.all()

    # ----------------------------------------------------------------------
    @property
    def length(self):
        """
        Returns the total length of the dataframe

        :returns: float

        >>> df.spatial.length
        1.23427

        """
        return self._data[self.name].geom.length.sum()

    # ----------------------------------------------------------------------
    def plot(self, map_widget: "MapView" = None, renderer: dict = None):
        """Displays the Dask DataFrame on a Map Widget"""
        return dask_plot(
            df=self._data,  # dask dataframe
            map_widget=map_widget,  # map view
            renderer=renderer or self.renderer,  # display information
        )

    # ----------------------------------------------------------------------
    def project(self, spatial_reference, transformation_name=None):
        """
        Reprojects the who dataset into a new spatial reference. This is an inplace operation meaning
        that it will update the defined geometry column from the `set_geometry`.

        **This requires ArcPy or pyproj v4**

        ====================     ====================================================================
        **Parameter**             **Description**
        --------------------     --------------------------------------------------------------------
        spatial_reference        Required SpatialReference. The new spatial reference. This can be a
                                 SpatialReference object or the coordinate system name.
        --------------------     --------------------------------------------------------------------
        transformation_name      Optional String. The geotransformation name.
        ====================     ====================================================================

        :returns: boolean
        """
        _HASARCPY, _HASSHAPELY, _HASPYPROJ = self._check_geometry_engine()
        try:
            if isinstance(spatial_reference, (int, str)) and _HASARCPY:
                import arcpy

                spatial_reference = arcpy.SpatialReference(spatial_reference)
                vals = (
                    self._data[self.name]
                    .geom.project_as(
                        **{
                            "spatial_reference": spatial_reference,
                            "transformation_name": transformation_name,
                        }
                    )
                    .compute()
                )
                self._data[self.name] = vals
                return True
            elif isinstance(spatial_reference, (int, str)) and _HASPYPROJ:
                vals = (
                    self._data[self.name]
                    .geom.project_as(
                        **{
                            "spatial_reference": spatial_reference,
                            "transformation_name": transformation_name,
                        }
                    )
                    .compute()
                )
                self._data[self.name] = vals
                return True
            else:
                return False
        except Exception as e:
            raise Exception(e)

    # ----------------------------------------------------------------------
    @lru_cache(maxsize=255)
    def _check_geometry_engine(self):
        _HASARCPY = True
        _HASSHAPELY = True
        _HASPYPROJ = True
        try:
            i = importlib.util.find_spec("arcpy")
            if i is None:
                raise ImportError("Cannot find arcpy.")
        except ImportError:
            _HASARCPY = False
        try:
            i = importlib.util.find_spec("shapely")
            if i is None:
                raise ImportError("Cannot find shapely.")
        except ImportError:
            _HASSHAPELY = False
        try:
            i = importlib.util.find_spec("pyproj")
            if i is None:
                raise ImportError("Cannot find pyproj.")
        except ImportError:
            _HASPYPROJ = False
        return _HASARCPY, _HASSHAPELY, _HASPYPROJ

    # ----------------------------------------------------------------------
    def sindex(self, reset=False) -> "DaskSpatialIndex":
        """
        Returns a dask specialized spatial index
        """

        if reset:
            self._sindex = None
        if self._sindex is None:
            results = (
                self._data.map_partitions(
                    lambda part: self._fn_method(
                        part.spatial, "sindex", **{"stype": "quadtree"}
                    )
                )
                .compute()
                .tolist()
            )
            self._sindex = DaskSpatialIndex(results)
        return self._sindex

    # ----------------------------------------------------------------------
    @property
    def renderer(self) -> "InsensitiveDict":
        """
        Define the renderer for the Dask DF.  If none is given, then the value is reset.

        :returns: InsensitiveDict
        """
        if self._renderer is None:
            self._renderer = self._build_renderer()
        return self._renderer

    # ----------------------------------------------------------------------
    @renderer.setter
    def renderer(self, renderer):
        """
        Define the renderer for the Dask DF.  If none is given, then the value is reset.

        :returns: InsensitiveDict
        """

        if renderer is None:
            renderer = self._build_renderer()
        if isinstance(renderer, dict):
            renderer = InsensitiveDict.from_dict(renderer)
        elif isinstance(renderer, PropertyMap):
            renderer = InsensitiveDict.from_dict(dict(renderer))
        elif isinstance(renderer, InsensitiveDict):
            pass
        else:
            raise ValueError("renderer must be a dictionary type.")
        self._renderer = renderer

    # ----------------------------------------------------------------------
    def set_geometry(self, col, sr=None, inplace=True):
        """Assigns the geometry column by name or by list

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        col                    Required string, Pandas Series, GeoArray, list or tuple. If a string, this
                               is the name of the column containing the geometry. If a Pandas Series
                               GeoArray, list or tuple, it is an iterable of Geometry objects.
        ------------------     --------------------------------------------------------------------
        sr                     Optional integer or spatial reference of the geometries described in
                               the first parameter. If the geometry objects already have the spatial
                               reference defined, this is not necessary. If the spatial reference for
                               the geometry objects is NOT define, it will default to WGS84 (wkid 4326).
        ------------------     --------------------------------------------------------------------
        inplace                Optional bool. Whether or not to modify the dataframe in place, or return
                               a new dataframe. If True, nothing is returned and the dataframe is modified
                               in place. If False, a new dataframe is returned with the geometry set.
                               Defaults to True.
        ==================     ====================================================================

        :return:
            Spatially Enabled DataFrame or None
        """
        return self._data.map_partitions(
            lambda part: self._fn_method(
                part.spatial,
                "set_geometry",
                **{"col": col, "sr": sr, "inplace": inplace},
            )
        )

    # ----------------------------------------------------------------------
    @property
    def sr(self):
        """returns the spatial references of the dataframe"""
        b = self._data.index.partitions[0]
        res = self._data.loc[b.to_series().compute().nlargest(5).tolist(), [self.name]][
            self.name
        ].geom.spatial_reference.unique()
        return res

    # ----------------------------------------------------------------------
    @staticmethod
    def from_featureclass(location: str, npartitions: int, **kwargs):
        """
        Returns a Spatially enabled `pandas.DataFrame` from a feature class.

        ===========================     ====================================================================
        **Parameter**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        location                        Required string or pathlib.Path. Full path to the feature class
        ---------------------------     --------------------------------------------------------------------
        npartitions                     Required int. The number of partitions of the index to create. Note
                                        that depending on the size and index of the dataframe, the output
                                        may have fewer partitions than requested.
        ===========================     ====================================================================

        *Optional parameters when ArcPy library is available in the current environment*:

        ===========================     ====================================================================
        **Optional Argument**           **Description**
        ---------------------------     --------------------------------------------------------------------
        sql_clause                      sql clause to parse data down. To learn more see
                                        `ArcPy Search Cursor <https://pro.arcgis.com/en/pro-app/arcpy/data-access/searchcursor-class.htm>`_
        ---------------------------     --------------------------------------------------------------------
        where_clause                    where statement. To learn more see `ArcPy SQL reference <https://pro.arcgis.com/en/pro-app/help/mapping/navigation/sql-reference-for-elements-used-in-query-expressions.htm>`_
        ---------------------------     --------------------------------------------------------------------
        fields                          list of strings specifying the field names.
        ---------------------------     --------------------------------------------------------------------
        spatial_filter                  A `Geometry` object that will filter the results.  This requires
                                        `arcpy` to work.
        ===========================     ====================================================================

        :returns: pandas.core.frame.DataFrame
        """
        return from_pandas(
            data=from_featureclass(filename=location, **kwargs),
            npartitions=npartitions,
        )

    # ----------------------------------------------------------------------
    def to_featureclass(
        self,
        location,
        overwrite=True,
        has_z=None,
        has_m=None,
        sanitize_columns=True,
    ):
        """
        Exports a dask dataframe to a feature class.

        ===========================     ====================================================================
        **Parameter**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        location                        Required string. The output of the table.
        ---------------------------     --------------------------------------------------------------------
        overwrite                       Optional Boolean.  If True and if the feature class exists, it will be
                                        deleted and overwritten.  This is default.  If False, the feature class
                                        and the feature class exists, and exception will be raised.
        ---------------------------     --------------------------------------------------------------------
        has_z                           Optional Boolean.  If True, the dataset will be forced to have Z
                                        based geometries.  If a geometry is missing a Z value when true, a
                                        RuntimeError will be raised.  When False, the API will not use the
                                        Z value.
        ---------------------------     --------------------------------------------------------------------
        has_m                           Optional Boolean.  If True, the dataset will be forced to have M
                                        based geometries.  If a geometry is missing a M value when true, a
                                        RuntimeError will be raised. When False, the API will not use the
                                        M value.
        ---------------------------     --------------------------------------------------------------------
        sanitize_columns                Optional Boolean. If True, column names will be converted to string,
                                        invalid characters removed and other checks will be performed. The
                                        default is True.
        ===========================     ====================================================================

        :returns: String

        """
        if location and not str(os.path.dirname(location)).lower() in [
            "memory",
            "in_memory",
        ]:
            location = os.path.abspath(path=location)
        df = self._data.compute().convert_dtypes()
        df.spatial.name
        return to_featureclass(
            geo=df.spatial,
            location=location,
            overwrite=overwrite,
            has_z=has_z,
            sanitize_columns=sanitize_columns,
            has_m=has_m,
        )

    # ----------------------------------------------------------------------
    def to_parquet(
        self,
        folder: str,
        index: bool = None,
        compression: str = "gzip",
        **kwargs,
    ):
        """
        Exports Each Dask DataFrame partition to a parquet file

        :returns: string (folder path)

        """
        if os.path.isdir(folder):
            os.makedirs(folder, exist_ok=True)
        results = []
        for i in range(self._data.npartitions):
            fp = os.path.join(f"{folder}", "part.%i.parquet" % (i))
            res = (
                self._data.get_partition(i)
                .compute()
                .spatial.to_parquet(fp, index=index, compression=compression, **kwargs)
            )
            results.append(res)
        return results

    # ----------------------------------------------------------------------
    @staticmethod
    def from_parquet(folder: str, ext: str = None):
        dfs = []
        from ._io._arrow import _read_parquet

        if ext is None:
            ext = "parquet"
        for fn in glob.glob(pathname=f"{folder}/*.{ext}"):
            dfs.append(_read_parquet(fn))
        df = pd.concat(dfs, ignore_index=True)
        df.reset_index(drop=True)
        return dd.from_pandas(df, 10)

    # ----------------------------------------------------------------------
    def _build_renderer(self):
        """sets the default symbology"""
        gt = self.geometry_type[0]
        base_renderer = {
            "labelingInfo": None,
            "label": "",
            "description": "",
            "type": "simple",
            "symbol": None,
        }
        if gt.lower() in ["point", "multipoint"]:
            base_renderer["symbol"] = {
                "color": [0, 128, 0, 128],
                "size": 18,
                "angle": 0,
                "xoffset": 0,
                "yoffset": 0,
                "type": "esriSMS",
                "style": "esriSMSCircle",
                "outline": {
                    "color": [0, 128, 0, 255],
                    "width": 1,
                    "type": "esriSLS",
                    "style": "esriSLSSolid",
                },
            }

        elif gt.lower() == "polyline":
            base_renderer["symbol"] = {
                "type": "esriSLS",
                "style": "esriSLSSolid",
                "color": [0, 128, 0, 128],
                "width": 1,
            }
        elif gt.lower() == "polygon":
            base_renderer["symbol"] = {
                "type": "esriSFS",
                "style": "esriSFSSolid",
                "color": [0, 128, 0, 128],
                "outline": {
                    "type": "esriSLS",
                    "style": "esriSLSSolid",
                    "color": [110, 110, 110, 255],
                    "width": 1,
                },
            }
        return InsensitiveDict(base_renderer)


###########################################################################
@register_series_accessor("geom")
class GeoDaskSeriesAccessor:
    """
    The Geospatial series accessor used to perform column specific operations.
    """

    _accessor_name = None
    _data = None
    _index = None
    _name = None

    # ----------------------------------------------------------------------
    def __init__(self, series: Series, *args, **kwargs):
        """initializer"""
        self._accessor_name = "geom"
        if series.dtype != "geometry":
            raise ValueError("'geometry' only works on Geometry dtypes.")
        self._data = series
        self._fn_attr = lambda a, op: getattr(a, op)
        self._fn_method = lambda a, op, **kwargs: getattr(a, op)(**kwargs)

    ##---------------------------------------------------------------------
    ##   Accessor Properties
    ##---------------------------------------------------------------------
    @property
    def area(self):
        """
        Returns the features area

        :returns: float in a series
        """
        return self._data.map_partitions(
            lambda part: self._fn_attr(part.geom, "area"),
            meta=pd.Series(dtype=float),
        )

    # ----------------------------------------------------------------------
    @property
    def as_arcpy(self):
        """
        Returns the features as ArcPy Geometry

        :returns: arcpy.Geometry in a series
        """
        return self._data.map_partitions(
            lambda part: self._fn_attr(part.geom, "as_arcpy"),
            meta=pd.Series(dtype=object),
        )

    # ----------------------------------------------------------------------
    @property
    def as_shapely(self):
        """
        Returns the features as Shapely Geometry

        :returns: shapely.Geometry in a series
        """
        return self._data.map_partitions(
            lambda part: self._fn_attr(part.geom, "as_shapely")
        )

    # ----------------------------------------------------------------------
    @property
    def centroid(self):
        """
        Returns the feature's centroid

        :returns: tuple (x,y) in series
        """
        return self._data.map_partitions(
            lambda part: self._fn_attr(part.geom, "centroid")
        )

    # ----------------------------------------------------------------------
    @property
    def extent(self):
        """
        Returns the feature's extent

        :returns: tuple (xmin,ymin,xmax,ymax) in series
        """
        return self._data.map_partitions(
            lambda part: self._fn_attr(part.geom, "extent")
        )

    # ----------------------------------------------------------------------
    @property
    def first_point(self):
        """
        Returns the feature's first point

        :returns: Geometry
        """
        return self._data.map_partitions(
            lambda part: self._fn_attr(part.geom, "first_point")
        )

    # ----------------------------------------------------------------------
    @property
    def geoextent(self):
        """
        A returns the geometry's extents

        :returns: Series of Floats
        """
        return self._data.map_partitions(
            lambda part: self._fn_attr(part.geom, "geoextent")
        )

    # ----------------------------------------------------------------------
    @property
    def geometry_type(self):
        """
        returns the geometry types

        :returns: Series of strings
        """
        return self._data.map_partitions(
            lambda part: self._fn_attr(part.geom, "geometry_type"),
            meta=pd.Series(dtype=str),
        )

    # ----------------------------------------------------------------------
    @property
    def hull_rectangle(self):
        """
        A space-delimited string of the coordinate pairs of the convex hull

        :returns: Series of strings
        """
        return self._data.map_partitions(
            lambda part: self._fn_attr(part.geom, "hull_rectangle")
        )

    # ----------------------------------------------------------------------
    @property
    def has_z(self):
        """
        Determines if the geometry has a Z value

        :returns: Series of Boolean
        """
        return self._data.map_partitions(
            lambda part: self._fn_attr(part.geom, "has_z"),
            meta=pd.Series(dtype=bool),
        )

    # ----------------------------------------------------------------------
    @property
    def has_m(self):
        """
        Determines if the geometry has a M value

        :returns: Series of Boolean
        """
        return self._data.map_partitions(
            lambda part: self._fn_attr(part.geom, "has_m"),
            meta=pd.Series(dtype=bool),
        )

    # ----------------------------------------------------------------------
    @property
    def is_empty(self):
        """
        Returns True/False if feature is empty

        :returns: Series of Booleans
        """
        return self._data.map_partitions(
            lambda part: self._fn_attr(part.geom, "is_empty"),
            meta=pd.Series(dtype=bool),
        )

    # ----------------------------------------------------------------------
    @property
    def is_multipart(self):
        """
        Returns True/False if features has multiple parts

        :returns: Series of Booleans
        """
        return self._data.map_partitions(
            lambda part: self._fn_attr(part.geom, "is_multipart"),
            meta=pd.Series(dtype=bool),
        )

    # ----------------------------------------------------------------------
    @property
    def is_valid(self):
        """
        Returns True/False if features geometry is valid

        :returns: Series of Booleans
        """
        return self._data.map_partitions(
            lambda part: self._fn_attr(part.geom, "is_valid"),
            meta=pd.Series(dtype=bool),
        )

    # ----------------------------------------------------------------------
    @property
    def JSON(self):
        """
        Returns JSON string  of Geometry

        :returns: Series of strings
        """
        return self._data.map_partitions(
            lambda part: self._fn_attr(part.geom, "JSON"),
            meta=pd.Series(dtype=str),
        )

    # ----------------------------------------------------------------------
    @property
    def label_point(self):
        """
        Returns the geometry point for the optimal label location

        :returns: Series of Geometries
        """
        return self._data.map_partitions(
            lambda part: self._fn_attr(part.geom, "label_point")
        )

    # ----------------------------------------------------------------------
    @property
    def last_point(self):
        """
        Returns the Geometry of the last point in a feature.

        :returns: Series of Geometry
        """
        return self._data.map_partitions(
            lambda part: self._fn_attr(part.geom, "last_point")
        )

    # ----------------------------------------------------------------------
    @property
    def length(self):
        """
        Returns the length of the features

        :returns: Series of float
        """
        return self._data.map_partitions(
            lambda part: self._fn_attr(part.geom, "length"),
            meta=pd.Series(dtype=float),
        )

    # ----------------------------------------------------------------------
    @property
    def length3D(self):
        """
        Returns the length of the features

        :returns: Series of float
        """
        return self._data.map_partitions(
            lambda part: self._fn_attr(part.geom, "length3D"),
            meta=pd.Series(dtype=float),
        )

    # ----------------------------------------------------------------------
    @property
    def part_count(self):
        """
        Returns the number of parts in a feature's geometry

        :returns: Series of Integer
        """
        return self._data.map_partitions(
            lambda part: self._fn_attr(part.geom, "part_count"),
            meta=pd.Series(dtype=int),
        )

    # ----------------------------------------------------------------------
    @property
    def point_count(self):
        """
        Returns the number of points in a feature's geometry

        :returns: Series of Integer
        """
        return self._data.map_partitions(
            lambda part: self._fn_attr(part.geom, "point_count"),
            meta=pd.Series(dtype=int),
        )

    # ----------------------------------------------------------------------
    @property
    def spatial_reference(self):
        """
        Returns the Spatial Reference of the Geometry

        :returns: Series of SpatialReference
        """
        return self._data.map_partitions(
            lambda part: self._fn_attr(part.geom, "spatial_reference")
        )

    # ----------------------------------------------------------------------
    @property
    def true_centroid(self):
        """
        Returns the true centroid of the Geometry

        :returns: Series of Points
        """
        return self._data.map_partitions(
            lambda part: self._fn_attr(part.geom, "true_centroid")
        )

    # ----------------------------------------------------------------------
    @property
    def WKB(self):
        """
        Returns the Geometry as WKB

        :returns: Series of Bytes
        """
        return self._data.map_partitions(lambda part: self._fn_attr(part.geom, "WKB"))

    # ----------------------------------------------------------------------
    @property
    def WKT(self):
        """
        Returns the Geometry as WKT

        :returns: Series of String
        """
        return self._data.map_partitions(
            lambda part: self._fn_attr(part.geom, "WKT"),
            meta=pd.Series(dtype=str),
        )

    ##---------------------------------------------------------------------
    ##  Accessor Geometry Method
    ##---------------------------------------------------------------------
    def vectorize(self, normalize=False):
        """
        Converts the Geometry to `numpy array` where the following holds true:

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        normalize           Optional Boolean.  If **True**, the values are normalized using a Min/Max Scalar.
        ===============     ====================================================================

        **Output index position of vectorized geometry**

        ==================     ====================================================================
        **Index Position**           **Description**
        ------------------     --------------------------------------------------------------------
        0                      Float. X Coordinate of the Geometry
        ------------------     --------------------------------------------------------------------
        1                      Float. Y Coordinate of the Geometry
        ------------------     --------------------------------------------------------------------
        2                      Integer. The values can be 0 or 1. Where 1 indicates the start of the a geometry.
        ------------------     --------------------------------------------------------------------
        3                      Integer. The values can be 0 or 1. Where 1 indicates Non endpoint or start point of geometry.
        ------------------     --------------------------------------------------------------------
        4                      Integer. The values can be 0 or 1. Where 1 indicates the endpoint of the a geometry.
        ------------------     --------------------------------------------------------------------
        5                      Integer. The values can be 0 or 1. Where 1 indicates the bounding geometery.
        ==================     ====================================================================

        **Geometry Output Example**

        >>> polygon.vectorize(normalize=False)
        np.array([[-2, 2, 1, 0, 0, 1],
                  [2, 2, 0, 1, 0, 1],
                  [1, 5, 0, 0, 1, 1]])


        This output describes a Polygon Geometry in the shape of a triangle with `normalize=False`

        :returns: Series of np.ndarray objects

        """
        pass

    def angle_distance_to(self, second_geometry: "Geometry", method: str = "GEODESIC"):
        """
        Returns a tuple of angle and distance to another point using a
        measurement type.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        second_geometry     Required Geometry.  A arcgis.Geometry object.
        ---------------     --------------------------------------------------------------------
        method              Optional String. PLANAR measurements reflect the projection of geographic
                            data onto the 2D surface (in other words, they will not take into
                            account the curvature of the earth). GEODESIC, GREAT_ELLIPTIC, and
                            LOXODROME measurement types may be chosen as
                            an alternative, if desired.
        ===============     ====================================================================

        :returns: a tuple of angle and distance to another point using a measurement type.
        """
        return self._data.map_partitions(
            lambda part: self._fn_method(
                part.geom,
                "angle_distance_to",
                **{"second_geometry": second_geometry, "method": method},
            )
        )

    # ----------------------------------------------------------------------
    def boundary(self):
        """
        Constructs the boundary of the geometry.

        :returns: arcgis.geometry.Polyline
        """
        return self._data.map_partitions(
            lambda part: self._fn_method(part.geom, "boundary")
        )

    # ----------------------------------------------------------------------
    def buffer(self, distance: float):
        """
        Constructs a polygon at a specified distance from the geometry.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        distance            Required float. The buffer distance. The buffer distance is in the
                            same units as the geometry that is being buffered.
                            A negative distance can only be specified against a polygon geometry.
        ===============     ====================================================================

        :returns: arcgis.geometry.Polygon
        """
        return self._data.map_partitions(
            lambda part: self._fn_method(part.geom, "buffer", **{"distance": distance})
        )

    # ----------------------------------------------------------------------
    def clip(self, envelope: tuple):
        """
        Constructs the intersection of the geometry and the specified extent.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        envelope            required tuple. The tuple must have (XMin, YMin, XMax, YMax) each value
                            represents the lower left bound and upper right bound of the extent.
        ===============     ====================================================================

        :returns: output geometry clipped to extent

        """
        return self._data.map_partitions(
            lambda part: self._fn_method(part.geom, "clip", **{"envelope": envelope})
        )

    # ----------------------------------------------------------------------
    def contains(self, second_geometry: "Geometry", relation: str = None):
        """
        Indicates if the base geometry contains the comparison geometry.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        second_geometry     Required arcgis.geometry.Geometry. A second geometry
        ---------------     --------------------------------------------------------------------
        relation            Optional string. The spatial relationship type.

                            + BOUNDARY - Relationship has no restrictions for interiors or boundaries.
                            + CLEMENTINI - Interiors of geometries must intersect. Specifying CLEMENTINI is equivalent to specifying None. This is the default.
                            + PROPER - Boundaries of geometries must not intersect.
        ===============     ====================================================================

        :returns: boolean
        """
        return self._data.map_partitions(
            lambda part: self._fn_method(
                part.geom,
                "contains",
                **{"second_geometry": second_geometry, "relation": relation},
            )
        )

    # ----------------------------------------------------------------------
    def convex_hull(self):
        """
        Constructs the geometry that is the minimal bounding polygon such
        that all outer angles are convex.
        """
        return self._data.map_partitions(
            lambda part: self._fn_method(part.geom, "convex_hull")
        )

    # ----------------------------------------------------------------------
    def crosses(self, second_geometry: "Geometry"):
        """
        Indicates if the two geometries intersect in a geometry of a lesser
        shape type.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        second_geometry     Required arcgis.geometry.Geometry. A second geometry
        ===============     ====================================================================

        :returns: boolean

        """
        return self._data.map_partitions(
            lambda part: self._fn_method(
                part.geom, "crosses", **{"second_geometry": second_geometry}
            )
        )

    # ----------------------------------------------------------------------
    def cut(self, cutter: "Polyline"):
        """
        Splits this geometry into a part left of the cutting polyline, and
        a part right of it.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        cutter              Required Polyline. The cuttin polyline geometry
        ===============     ====================================================================

        :returns: a list of two geometries

        """
        return self._data.map_partitions(
            lambda part: self._fn_method(part.geom, "cut", **{"cutter": cutter})
        )

    # ----------------------------------------------------------------------
    def densify(self, method: str, distance: float, deviation: float):
        """
        Creates a new geometry with added vertices

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        method              Required String. The type of densification, DISTANCE, ANGLE, or GEODESIC
        ---------------     --------------------------------------------------------------------
        distance            Required float. The maximum distance between vertices. The actual
                            distance between vertices will usually be less than the maximum
                            distance as new vertices will be evenly distributed along the
                            original segment. If using a type of DISTANCE or ANGLE, the
                            distance is measured in the units of the geometry's spatial
                            reference. If using a type of GEODESIC, the distance is measured
                            in meters.
        ---------------     --------------------------------------------------------------------
        deviation           Required float. Densify uses straight lines to approximate curves.
                            You use deviation to control the accuracy of this approximation.
                            The deviation is the maximum distance between the new segment and
                            the original curve. The smaller its value, the more segments will
                            be required to approximate the curve.
        ===============     ====================================================================

        :returns: arcgis.geometry.Geometry

        """
        return self._data.map_partitions(
            lambda part: self._fn_method(
                part.geom,
                "densify",
                **{
                    "method": method,
                    "distance": distance,
                    "deviation": deviation,
                },
            )
        )

    # ----------------------------------------------------------------------
    def difference(self, second_geometry: "Geometry"):
        """
        Constructs the geometry that is composed only of the region unique
        to the base geometry but not part of the other geometry. The
        following illustration shows the results when the red polygon is the
        source geometry.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        second_geometry     Required arcgis.geometry.Geometry. A second geometry
        ===============     ====================================================================

        :returns: arcgis.geometry.Geometry

        """
        return self._data.map_partitions(
            lambda part: self._fn_method(
                part.geom,
                "difference",
                **{"second_geometry": second_geometry},
            )
        )

    # ----------------------------------------------------------------------
    def disjoint(self, second_geometry: "Geometry"):
        """
        Indicates if the base and comparison geometries share no points in
        common.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        second_geometry     Required arcgis.geometry.Geometry. A second geometry
        ===============     ====================================================================

        :returns: boolean

        """
        return self._data.map_partitions(
            lambda part: self._fn_method(
                part.geom, "disjoin", **{"second_geometry": second_geometry}
            )
        )

    # ----------------------------------------------------------------------
    def distance_to(self, second_geometry: "Geometry"):
        """
        Returns the minimum distance between two geometries. If the
        geometries intersect, the minimum distance is 0.
        Both geometries must have the same projection.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        second_geometry     Required arcgis.geometry.Geometry. A second geometry
        ===============     ====================================================================

        :returns: float

        """
        return self._data.map_partitions(
            lambda part: self._fn_method(
                part.geom,
                "distance_to",
                **{"second_geometry": second_geometry},
            )
        )

    # ----------------------------------------------------------------------
    def equals(self, second_geometry: "Geometry"):
        """
        Indicates if the base and comparison geometries are of the same
        shape type and define the same set of points in the plane. This is
        a 2D comparison only; M and Z values are ignored.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        second_geometry     Required arcgis.geometry.Geometry. A second geometry
        ===============     ====================================================================

        :returns: boolean


        """
        return self._data.map_partitions(
            lambda part: self._fn_method(
                part.geom, "equals", **{"second_geometry": second_geometry}
            )
        )

    # ----------------------------------------------------------------------
    def generalize(self, max_offset: float):
        """
        Creates a new simplified geometry using a specified maximum offset
        tolerance.  This only works on Polylines and Polygons.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        max_offset          Required float. The maximum offset tolerance.
        ===============     ====================================================================

        :returns: arcgis.geometry.Geometry

        """
        return self._data.map_partitions(
            lambda part: self._fn_method(
                part.geom, "generalize", **{"max_offset": max_offset}
            )
        )

    # ----------------------------------------------------------------------
    def get_area(self, method: str, units: str = None):
        """
        Returns the area of the feature using a measurement type.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        method              Required String. PLANAR measurements reflect the projection of
                            geographic data onto the 2D surface (in other words, they will not
                            take into account the curvature of the earth). GEODESIC,
                            GREAT_ELLIPTIC, LOXODROME, and PRESERVE_SHAPE measurement types
                            may be chosen as an alternative, if desired.
        ---------------     --------------------------------------------------------------------
        units               Optional String. Areal unit of measure keywords: ACRES | ARES | HECTARES
                            | SQUARECENTIMETERS | SQUAREDECIMETERS | SQUAREINCHES | SQUAREFEET
                            | SQUAREKILOMETERS | SQUAREMETERS | SQUAREMILES |
                            SQUAREMILLIMETERS | SQUAREYARDS
        ===============     ====================================================================

        :returns: float

        """
        return self._data.map_partitions(
            lambda part: self._fn_method(
                part.geom, "get_area", **{"method": method, "units": units}
            )
        )

    # ----------------------------------------------------------------------
    def get_length(self, method: str, units: str):
        """
        Returns the length of the feature using a measurement type.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        method              Required String. PLANAR measurements reflect the projection of
                            geographic data onto the 2D surface (in other words, they will not
                            take into account the curvature of the earth). GEODESIC,
                            GREAT_ELLIPTIC, LOXODROME, and PRESERVE_SHAPE measurement types
                            may be chosen as an alternative, if desired.
        ---------------     --------------------------------------------------------------------
        units               Required String. Linear unit of measure keywords: CENTIMETERS |
                            DECIMETERS | FEET | INCHES | KILOMETERS | METERS | MILES |
                            MILLIMETERS | NAUTICALMILES | YARDS
        ===============     ====================================================================

        :returns: float

        """
        return self._data.map_partitions(
            lambda part: self._fn_method(
                part.geom, "get_length", **{"method": method, "units": units}
            )
        )

    # ----------------------------------------------------------------------
    def get_part(self, index: int = None):
        """
        Returns an array of point objects for a particular part of geometry
        or an array containing a number of arrays, one for each part.

        **requires arcpy**

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        index               Required Integer. The index position of the geometry.
        ===============     ====================================================================

        :return: arcpy.Array

        """
        return self._data.map_partitions(
            lambda part: self._fn_method(part.geom, "get_part", **{"index": index})
        )

    # ----------------------------------------------------------------------
    def intersect(self, second_geometry: "Geometry", dimension: int = 1):
        """
        Constructs a geometry that is the geometric intersection of the two
        input geometries. Different dimension values can be used to create
        different shape types. The intersection of two geometries of the
        same shape type is a geometry containing only the regions of overlap
        between the original geometries.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        second_geometry     Required arcgis.geometry.Geometry. A second geometry
        ---------------     --------------------------------------------------------------------
        dimension           Required Integer. The topological dimension (shape type) of the
                            resulting geometry.

                            + 1  -A zero-dimensional geometry (point or multipoint).
                            + 2  -A one-dimensional geometry (polyline).
                            + 4  -A two-dimensional geometry (polygon).

        ===============     ====================================================================

        :returns: boolean

        """
        return self._data.map_partitions(
            lambda part: self._fn_method(
                part.geom,
                "intersect",
                **{
                    "second_geometry": second_geometry,
                    "dimension": dimension,
                },
            )
        )

    # ----------------------------------------------------------------------
    def measure_on_line(self, second_geometry: "Geometry", as_percentage: bool = False):
        """
        Returns a measure from the start point of this line to the in_point.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        second_geometry     Required arcgis.geometry.Geometry. A second geometry
        ---------------     --------------------------------------------------------------------
        as_percentage       Optional Boolean. If False, the measure will be returned as a
                            distance; if True, the measure will be returned as a percentage.
        ===============     ====================================================================

        :return: float

        """
        return self._data.map_partitions(
            lambda part: self._fn_method(
                part.geom,
                "measure_on_line",
                **{
                    "second_geometry": second_geometry,
                    "as_percentage": as_percentage,
                },
            )
        )

    # ----------------------------------------------------------------------
    def overlaps(self, second_geometry: "Geometry"):
        """
        Indicates if the intersection of the two geometries has the same
        shape type as one of the input geometries and is not equivalent to
        either of the input geometries.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        second_geometry     Required arcgis.geometry.Geometry. A second geometry
        ===============     ====================================================================

        :return: boolean

        """
        return self._data.map_partitions(
            lambda part: self._fn_method(
                part.geom, "overlaps", **{"second_geometry": second_geometry}
            )
        )

    # ----------------------------------------------------------------------
    def point_from_angle_and_distance(
        self, angle: float, distance: float, method: str = "GEODESCIC"
    ):
        """
        Returns a point at a given angle and distance in degrees and meters
        using the specified measurement type.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        angle               Required Float. The angle in degrees to the returned point.
        ---------------     --------------------------------------------------------------------
        distance            Required Float. The distance in meters to the returned point.
        ---------------     --------------------------------------------------------------------
        method              Optional String. PLANAR measurements reflect the projection of geographic
                            data onto the 2D surface (in other words, they will not take into
                            account the curvature of the earth). GEODESIC, GREAT_ELLIPTIC,
                            LOXODROME, and PRESERVE_SHAPE measurement types may be chosen as
                            an alternative, if desired.
        ===============     ====================================================================

        :return: arcgis.geometry.Geometry


        """
        return self._data.map_partitions(
            lambda part: self._fn_method(
                part.geom,
                "point_from_angle_and_distance",
                **{"angle": angle, "distance": distance, "method": method},
            )
        )

    # ----------------------------------------------------------------------
    def position_along_line(self, value: float, use_percentage: bool = False):
        """
        Returns a point on a line at a specified distance from the beginning
        of the line.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        value               Required Float. The distance along the line.
        ---------------     --------------------------------------------------------------------
        use_percentage      Optional Boolean. The distance may be specified as a fixed unit
                            of measure or a ratio of the length of the line. If True, value
                            is used as a percentage; if False, value is used as a distance.
                            For percentages, the value should be expressed as a double from
                            0.0 (0%) to 1.0 (100%).
        ===============     ====================================================================

        :return: Geometry

        """
        return self._data.map_partitions(
            lambda part: self._fn_method(
                part.geom,
                "position_along_line",
                **{"value": value, "use_percentage": use_percentage},
            )
        )

    # ----------------------------------------------------------------------
    def project_as(
        self,
        spatial_reference: "SpatialReference",
        transformation_name: str = None,
    ):
        """
        Projects a geometry and optionally applies a geotransformation.

        ====================     ====================================================================
        **Parameter**             **Description**
        --------------------     --------------------------------------------------------------------
        spatial_reference        Required SpatialReference. The new spatial reference. This can be a
                                 SpatialReference object or the coordinate system name.
        --------------------     --------------------------------------------------------------------
        transformation_name      Required String. The geotransformation name.
        ====================     ====================================================================

        :returns: arcgis.geometry.Geometry
        """
        return self._data.map_partitions(
            lambda part: self._fn_method(
                part.geom,
                "project_as",
                **{
                    "spatial_reference": spatial_reference,
                    "transformation_name": transformation_name,
                },
            )
        )

    # ----------------------------------------------------------------------
    def query_point_and_distance(
        self, second_geometry: "Geometry", use_percentage: bool = False
    ):
        """
        Finds the point on the polyline nearest to the in_point and the
        distance between those points. Also returns information about the
        side of the line the in_point is on as well as the distance along
        the line where the nearest point occurs.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        second_geometry     Required arcgis.geometry.Geometry. A second geometry
        ---------------     --------------------------------------------------------------------
        as_percentage       Optional boolean - if False, the measure will be returned as
                            distance, True, measure will be a percentage
        ===============     ====================================================================

        :return: tuple

        """
        return self._data.map_partitions(
            lambda part: self._fn_method(
                part.geom,
                "query_point_and_distance",
                **{
                    "second_geometry": second_geometry,
                    "use_percentage": use_percentage,
                },
            )
        )

    # ----------------------------------------------------------------------
    def segment_along_line(
        self,
        start_measure: float,
        end_measure: float,
        use_percentage: bool = False,
    ):
        """
        Returns a Polyline between start and end measures. Similar to
        Polyline.positionAlongLine but will return a polyline segment between
        two points on the polyline instead of a single point.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        start_measure       Required Float. The starting distance from the beginning of the line.
        ---------------     --------------------------------------------------------------------
        end_measure         Required Float. The ending distance from the beginning of the line.
        ---------------     --------------------------------------------------------------------
        use_percentage      Optional Boolean. The start and end measures may be specified as
                            fixed units or as a ratio.
                            If True, start_measure and end_measure are used as a percentage; if
                            False, start_measure and end_measure are used as a distance. For
                            percentages, the measures should be expressed as a double from 0.0
                            (0 percent) to 1.0 (100 percent).
        ===============     ====================================================================

        :returns: Geometry

        """
        return self._data.map_partitions(
            lambda part: self._fn_method(
                part.geom,
                "segment_along_line",
                **{
                    "start_measure": start_measure,
                    "end_measure": end_measure,
                    "use_percentage": use_percentage,
                },
            )
        )

    # ----------------------------------------------------------------------
    def snap_to_line(self, second_geometry: "Geometry"):
        """
        Returns a new point based on in_point snapped to this geometry.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        second_geometry     Required arcgis.geometry.Geometry. A second geometry
        ===============     ====================================================================

        :return: arcgis.gis.Geometry

        """
        return self._data.map_partitions(
            lambda part: self._fn_method(
                part.geom,
                "snap_to_line",
                **{"second_geometry": second_geometry},
            )
        )

    # ----------------------------------------------------------------------
    def symmetric_difference(self, second_geometry: "Geometry"):
        """
        Constructs the geometry that is the union of two geometries minus the
        instersection of those geometries.

        The two input geometries must be the same shape type.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        second_geometry     Required arcgis.geometry.Geometry. A second geometry
        ===============     ====================================================================

        :return: arcgis.gis.Geometry
        """
        return self._data.map_partitions(
            lambda part: self._fn_method(
                part.geom,
                "symmetric_difference",
                **{"second_geometry": second_geometry},
            )
        )

    # ----------------------------------------------------------------------
    def touches(self, second_geometry: "Geometry"):
        """
        Indicates if the boundaries of the geometries intersect.


        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        second_geometry     Required arcgis.geometry.Geometry. A second geometry
        ===============     ====================================================================

        :return: boolean
        """
        return self._data.map_partitions(
            lambda part: self._fn_method(
                part.geom, "touches", **{"second_geometry": second_geometry}
            )
        )

    # ----------------------------------------------------------------------
    def union(self, second_geometry: "Geometry"):
        """
        Constructs the geometry that is the set-theoretic union of the input
        geometries.


        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        second_geometry     Required arcgis.geometry.Geometry. A second geometry
        ===============     ====================================================================

        :return: arcgis.gis.Geometry
        """
        return self._data.map_partitions(
            lambda part: self._fn_method(
                part.geom, "union", **{"second_geometry": second_geometry}
            )
        )

    # ----------------------------------------------------------------------
    def within(self, second_geometry: "Geometry", relation: str = None):
        """
        Indicates if the base geometry is within the comparison geometry.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        second_geometry     Required arcgis.geometry.Geometry. A second geometry
        ---------------     --------------------------------------------------------------------
        relation            Optional String. The spatial relationship type.

                            - BOUNDARY  - Relationship has no restrictions for interiors or boundaries.
                            - CLEMENTINI  - Interiors of geometries must intersect. Specifying CLEMENTINI is equivalent to specifying None. This is the default.
                            - PROPER  - Boundaries of geometries must not intersect.

        ===============     ====================================================================

        :return: boolean

        """
        return self._data.map_partitions(
            lambda part: self._fn_method(
                part.geom,
                "within",
                **{"second_geometry": second_geometry, "relation": relation},
            )
        )


# -------------------------------------------------------------------------
@make_array_nonempty.register(GeoType)
def _(dtype):
    a = np.array([_default_geometry(), _default_geometry()], dtype=GeoType)
    return _from_geometry(a)


# -------------------------------------------------------------------------
@make_scalar.register(GeoType.type)
def _(x):
    return _default_geometry()


# -------------------------------------------------------------------------
@make_meta_dispatch.register(Geometry)
def make_meta_arcgis_geometry(x, index=None):
    return x


@normalize_token.register(GeoArray)
def tokenize_geometryarray(x):
    return uuid.uuid4().hex



from __future__ import annotations
import re
import urllib.parse as urllib_parse
from functools import lru_cache
from requests import Response
from typing import Any
from .._error import EsriHttpResponseError

__all__ = [
    "parse_url",
    "_split_username",
    "assemble_url",
    "check_response_for_error",
]


# @lru_cache(maxsize=255)
def check_response_for_error(resp: Response, raise_error: bool = False) -> str | None:
    """
    Checks the payload for `error` keyword.

    ===============     ====================================================================
    **Parameter**        **Description**
    ---------------     --------------------------------------------------------------------
    resp                requests.Response. The response object to examin.  First the
                        response object is checked to see if it's JSON, then it examines the
                        response for `error` keywords.
    ---------------     --------------------------------------------------------------------
    raise_error         boolean. If True and an error is found, the  EsriHttpResponseError
                        is raised, else the message is returned as a string.
    ===============     ====================================================================

    :returns: str

    :raise: EsriHttpResponseError
    """
    data: dict[str, Any]
    try:
        data = resp.json()
    except Exception as ex:
        return str(ex)
    if "error" in data and "message" in data["error"] and raise_error:
        raise EsriHttpResponseError(data["message"]["error"])
    elif "error" in data and "message" in data["error"] and raise_error == False:
        return data["error"]["message"]
    return None


@lru_cache(maxsize=255)
def parse_url(url: str) -> object:
    """
    Parses a URL string into it's pieces.

    :returns: Named Tuple
    """
    return urllib_parse.urlparse(url)


@lru_cache(maxsize=255)
def assemble_url(parsed: object) -> str:
    """
    creates the URL from a parsed URL
    """
    if parsed.port:
        netloc: str = parsed.netloc.split(":")[0]
        server_url = (
            f'{parsed.scheme}://{netloc}:{parsed.port}/{parsed.path[1:].split("/")[0]}'
        )
    else:
        server_url = (
            f'{parsed.scheme}://{parsed.netloc}/{parsed.path[1:].split("/")[0]}'
        )
    return server_url


@lru_cache(maxsize=255)
def _split_username(username: str) -> list[str]:
    regex = r"(\S*)?(@|#|//|\\\\|(?<!/)/(?!/)|\\)(\S*)"
    matches = re.finditer(regex, username, re.IGNORECASE | re.DOTALL)
    tokens = []
    for match in matches:
        for group in match.groups():
            tokens.append(group)

    if tokens[1] in ["//", "\\", "/"]:
        uname = tokens[2]
        dom = tokens[0]
    elif tokens[1] == "@":
        uname = tokens[0]
        dom = tokens[2]

    return [uname, dom]
